# Copyright 2021 Canonical Ltd.
# See LICENSE file for licensing details.
"""Types used by cos-lib."""
from typing import Dict, List, Literal, Optional, TypedDict, Union

from typing_extensions import Required

QueryType = Literal["logql", "promql"]
RuleType = Literal["alert", "record"]


class _RecordingRuleFormat(TypedDict):
    record: Required[str]
    expr: Required[str]
    labels: Dict[str, str]


class _AlertingRuleFormat(TypedDict):
    alert: Required[str]
    expr: Required[str]
    duration: Optional[str]
    keep_firing_for: Optional[str]
    labels: Dict[str, str]
    annotations: Optional[Dict[str, str]]


SingleRuleFormat = Union[_AlertingRuleFormat, _RecordingRuleFormat]


class OfficialRuleFileItem(TypedDict):
    """Typing for a single node of the official rule file format."""

    name: str
    rules: List[SingleRuleFormat]


class OfficialRuleFileFormat(TypedDict):
    """Typing for the official rule file format.

    References:
    - https://prometheus.io/docs/prometheus/latest/configuration/alerting_rules/
    - https://prometheus.io/docs/prometheus/latest/configuration/recording_rules/
    """

    groups: List[OfficialRuleFileItem]



import dataclasses
from unittest.mock import MagicMock, PropertyMock, patch

import httpx
import ops
import pytest
import tenacity
from lightkube import ApiError
from ops import testing

from cosl.coordinated_workers.coordinator import ClusterRolesConfig, Coordinator
from cosl.interfaces.cluster import ClusterProviderAppData, ClusterRequirerAppData
from tests.test_coordinated_workers.test_worker_status import k8s_patch

my_roles = ClusterRolesConfig(
    roles={"role"},
    meta_roles={},
    minimal_deployment={"role": 1},
    recommended_deployment={"role": 2},
)


class MyCoordCharm(ops.CharmBase):
    def __init__(self, framework: ops.Framework):
        super().__init__(framework)

        self.coordinator = Coordinator(
            charm=self,
            roles_config=my_roles,
            external_url="localhost:3200",
            worker_metrics_port="8080",
            endpoints={
                "cluster": "cluster",
                "s3": "s3",
                "certificates": "certificates",
                "grafana-dashboards": "grafana-dashboard",
                "logging": "logging",
                "metrics": "metrics-endpoint",
                "charm-tracing": "self-charm-tracing",
                "workload-tracing": "self-workload-tracing",
                "send-datasource": None,
                "receive-datasource": "my-ds-exchange-require",
            },
            nginx_config=lambda _: "nginx config",
            workers_config=lambda _: "worker config",
            resources_requests=lambda _: {"cpu": "50m", "memory": "100Mi"},
            container_name="charm",
        )


@pytest.fixture
def coord_charm():
    with k8s_patch():
        yield MyCoordCharm


@pytest.fixture
def ctx(coord_charm):
    return testing.Context(
        coord_charm,
        meta={
            "name": "lilith",
            "requires": {
                "s3": {"interface": "s3"},
                "logging": {"interface": "loki_push_api"},
                "certificates": {"interface": "tls-certificates"},
                "self-charm-tracing": {"interface": "tracing", "limit": 1},
                "self-workload-tracing": {"interface": "tracing", "limit": 1},
                "my-ds-exchange-require": {"interface": "grafana_datasource_exchange"},
            },
            "provides": {
                "cluster": {"interface": "cluster"},
                "grafana-dashboard": {"interface": "grafana_dashboard"},
                "metrics-endpoint": {"interface": "prometheus_scrape"},
                "my-ds-exchange-provide": {"interface": "grafana_datasource_exchange"},
            },
            "containers": {
                "nginx": {"type": "oci-image"},
                "nginx-prometheus-exporter": {"type": "oci-image"},
            },
        },
    )


@pytest.fixture()
def s3():
    return testing.Relation(
        "s3",
        remote_app_data={
            "access-key": "key",
            "bucket": "tempo",
            "endpoint": "http://1.2.3.4:9000",
            "secret-key": "soverysecret",
        },
        local_unit_data={"bucket": "tempo"},
    )


@pytest.fixture()
def worker():
    app_data = {}
    ClusterProviderAppData(worker_config="some: yaml").dump(app_data)
    remote_app_data = {}
    ClusterRequirerAppData(role="role").dump(remote_app_data)
    return testing.Relation("cluster", local_app_data=app_data, remote_app_data=remote_app_data)


@pytest.fixture()
def base_state(s3, worker):
    return testing.State(
        leader=True,
        containers={testing.Container("nginx"), testing.Container("nginx-prometheus-exporter")},
        relations={worker, s3},
    )


def set_containers(state, nginx_can_connect=False, exporter_can_connect=False):
    containers = {
        testing.Container("nginx", can_connect=nginx_can_connect),
        testing.Container("nginx-prometheus-exporter", can_connect=exporter_can_connect),
    }
    return dataclasses.replace(state, containers=containers)


@patch(
    "charms.observability_libs.v0.kubernetes_compute_resources_patch.ResourcePatcher.apply",
    MagicMock(return_value=None),
)
def test_status_check_no_workers(ctx, base_state, s3, caplog):
    # GIVEN the container cannot connect
    state = set_containers(base_state, True, False)
    state = dataclasses.replace(state, relations={s3})

    # WHEN we run any event
    state_out = ctx.run(ctx.on.config_changed(), state)

    # THEN the charm sets blocked
    assert state_out.unit_status == ops.BlockedStatus("[consistency] Missing any worker relation.")


@patch(
    "charms.observability_libs.v0.kubernetes_compute_resources_patch.ResourcePatcher.apply",
    MagicMock(return_value=None),
)
def test_status_check_no_s3(ctx, base_state, worker, caplog):
    # GIVEN the container cannot connect
    state = set_containers(base_state, True, False)
    state = dataclasses.replace(base_state, relations={worker})

    # WHEN we run any event
    state_out = ctx.run(ctx.on.config_changed(), state)

    # THEN the charm sets blocked
    assert state_out.unit_status == ops.BlockedStatus("[s3] Missing S3 integration.")


@patch(
    "charms.observability_libs.v0.kubernetes_compute_resources_patch.KubernetesComputeResourcesPatch.get_status",
    MagicMock(return_value=(ops.BlockedStatus(""))),
)
def test_status_check_k8s_patch_failed(ctx, base_state, caplog):
    # GIVEN the container can connect
    state = set_containers(base_state, True, True)

    # WHEN we run any event
    state_out = ctx.run(ctx.on.update_status(), state)

    assert state_out.unit_status == ops.BlockedStatus("")


@patch("charms.observability_libs.v0.kubernetes_compute_resources_patch.ResourcePatcher")
@patch(
    "cosl.coordinated_workers.coordinator.KubernetesComputeResourcesPatch.PATCH_RETRY_STOP",
    PropertyMock(return_value=tenacity.wait_fixed(1)),
)
def test_status_check_k8s_patch_success_after_retries(
    resource_patcher_mock, ctx, base_state, caplog
):
    # GIVEN the container can connect
    state = set_containers(base_state, True, True)

    # Retry on that error
    response = httpx.Response(
        status_code=404, content='{"status": {"code": 404, "message": "Not Found"},"code":"404"}'
    )
    # Success on 2nd try
    resource_patcher_mock.return_value.apply.side_effect = [ApiError(response=response), None]

    # on collect-unit-status, the request patches are not yet reflected
    with patch(
        "cosl.coordinated_workers.coordinator.KubernetesComputeResourcesPatch.get_status",
        MagicMock(return_value=ops.WaitingStatus("waiting")),
    ):
        state_intermediate = ctx.run(ctx.on.config_changed(), state)
    assert state_intermediate.unit_status == ops.WaitingStatus("waiting")

    with patch(
        "cosl.coordinated_workers.coordinator.KubernetesComputeResourcesPatch.get_status",
        MagicMock(return_value=ops.ActiveStatus("")),
    ):
        state_out = ctx.run(ctx.on.update_status(), state_intermediate)
    assert state_out.unit_status == ops.ActiveStatus("Degraded.")



from cosl.interfaces.utils import DatabagModel


def test_databag_dump_update():
    class MyModel(DatabagModel):
        foo: int
        bar: str = "barian"

    db = {}
    assert MyModel(foo=1, bar="soz").dump(db)
    assert db == {"foo": "1", "bar": '"soz"'}


def test_databag_dumps():
    class MyModel(DatabagModel):
        foo: int
        bar: str = "barian"

    assert MyModel(foo=1, bar="bearian").dump() == {"foo": "1", "bar": '"bearian"'}



import dataclasses
from contextlib import ExitStack, contextmanager
from functools import partial
from unittest.mock import MagicMock, patch

import ops
import pytest
import tenacity
from ops import testing

from cosl.coordinated_workers.worker import (
    NoReadinessCheckEndpointConfiguredError,
    Worker,
)
from cosl.interfaces.cluster import ClusterProviderAppData


@pytest.fixture(params=[True, False])
def tls(request):
    return request.param


@contextmanager
def _urlopen_patch(url: str, resp: str, tls: bool):
    if url == f"{'https' if tls else 'http'}://localhost:3200/ready":
        mm = MagicMock()
        mm.read = MagicMock(return_value=resp.encode("utf-8"))
        yield mm
    else:
        raise RuntimeError("unknown path")


@contextmanager
def k8s_patch(status=ops.ActiveStatus(), is_ready=True):
    with patch("lightkube.core.client.GenericSyncClient"):
        with patch.multiple(
            "cosl.coordinated_workers.worker.KubernetesComputeResourcesPatch",
            _namespace="test-namespace",
            _patch=MagicMock(return_value=None),
            get_status=MagicMock(return_value=status),
            is_ready=MagicMock(return_value=is_ready),
        ) as patcher:
            yield patcher


@pytest.fixture(autouse=True)
def patch_status_wait():
    with ExitStack() as stack:
        # so we don't have to wait for minutes:
        stack.enter_context(
            patch(
                "cosl.coordinated_workers.worker.Worker.SERVICE_STATUS_UP_RETRY_WAIT",
                new=tenacity.wait_none(),
            )
        )
        stack.enter_context(
            patch(
                "cosl.coordinated_workers.worker.Worker.SERVICE_STATUS_UP_RETRY_STOP",
                new=tenacity.stop_after_delay(2),
            )
        )


@pytest.fixture
def ctx(tls):
    class MyCharm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self.worker = Worker(
                self,
                "workload",
                lambda _: ops.pebble.Layer(
                    {
                        "summary": "summary",
                        "services": {"service": {"summary": "summary", "override": "replace"}},
                    }
                ),
                {"cluster": "cluster"},
                readiness_check_endpoint=self._readiness_check_endpoint,
                resources_requests=lambda _: {"cpu": "50m", "memory": "100Mi"},
                container_name="charm",
            )

        def _readiness_check_endpoint(self, _):
            return f"{'https' if tls else 'http'}://localhost:3200/ready"

    return testing.Context(
        MyCharm,
        meta={
            "name": "lilith",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"workload": {"type": "oci-image"}},
        },
        config={
            "options": {
                "role-all": {"type": "bool", "default": False},
                "role-read": {"type": "bool", "default": True},
                "role-write": {"type": "bool", "default": True},
            }
        },
    )


@pytest.fixture(params=[True, False])
def base_state(request):
    app_data = {}
    ClusterProviderAppData(worker_config="some: yaml").dump(app_data)
    return testing.State(
        leader=request.param,
        containers={
            testing.Container(
                "workload",
                can_connect=True,
                execs={testing.Exec(("update-ca-certificates", "--fresh"))},
            )
        },
        relations={testing.Relation("cluster", remote_app_data=app_data)},
    )


@contextmanager
def endpoint_starting(tls):
    with patch(
        "urllib.request.urlopen",
        new=partial(_urlopen_patch, tls=tls, resp="foo\nStarting: 10\n bar"),
    ):
        yield


@contextmanager
def endpoint_ready(tls):
    with patch("urllib.request.urlopen", new=partial(_urlopen_patch, tls=tls, resp="ready")):
        yield


@contextmanager
def config_on_disk():
    with patch(
        "cosl.coordinated_workers.worker.Worker._running_worker_config", new=lambda _: True
    ):
        yield


@k8s_patch()
def test_status_check_no_pebble(ctx, base_state, caplog):
    # GIVEN the container cannot connect
    state = dataclasses.replace(
        base_state, containers={testing.Container("workload", can_connect=False)}
    )

    # WHEN we run any event
    state_out = ctx.run(ctx.on.update_status(), state)

    # THEN the charm sets blocked
    assert state_out.unit_status == ops.WaitingStatus("Waiting for `workload` container")


@k8s_patch()
def test_status_check_no_config(ctx, base_state, caplog):
    # GIVEN there is no config file on disk
    # WHEN we run any event
    with patch(
        "cosl.coordinated_workers.worker.Worker._running_worker_config", new=lambda _: None
    ):
        state_out = ctx.run(ctx.on.update_status(), base_state)

    # THEN the charm sets blocked
    assert state_out.unit_status == ops.WaitingStatus(
        "Waiting for coordinator to publish a config"
    )
    # AND THEN the charm logs that the config isn't on disk
    assert "Config file not on disk. Skipping status check." in caplog.messages


@k8s_patch()
def test_status_check_starting(ctx, base_state, tls):
    # GIVEN getting the status returns "Starting: X"
    with endpoint_starting(tls):
        # AND GIVEN that the config is on disk
        with config_on_disk():
            # AND GIVEN that the container can connect (default in base_state)
            state = base_state
            # WHEN we run any event
            state_out = ctx.run(ctx.on.update_status(), state)
    # THEN the charm sets waiting: Starting...
    assert state_out.unit_status == ops.WaitingStatus("Starting...")


@k8s_patch()
def test_status_check_ready(ctx, base_state, tls):
    # GIVEN getting the status returns "ready"
    with endpoint_ready(tls):
        # AND GIVEN that the config is on disk
        with config_on_disk():
            # AND GIVEN that the container can connect
            state = base_state
            # WHEN we run any event
            state_out = ctx.run(ctx.on.update_status(), state)
    # THEN the charm sets waiting: Starting...
    assert state_out.unit_status == ops.ActiveStatus("read,write ready.")


def test_status_no_endpoint(ctx, base_state, caplog):
    # GIVEN a charm doesn't pass an endpoint to Worker
    class MyCharm(ops.CharmBase):
        def __init__(self, framework: ops.Framework):
            super().__init__(framework)
            self.worker = Worker(
                self,
                "workload",
                lambda _: ops.pebble.Layer({"services": {"foo": {"command": "foo"}}}),
                {"cluster": "cluster"},
            )

    ctx = testing.Context(
        MyCharm,
        meta={
            "name": "damian",
            "requires": {"cluster": {"interface": "cluster"}},
            "containers": {"workload": {"type": "oci-image"}},
        },
        config={
            "options": {
                "role-all": {"type": "bool", "default": False},
                "role-read": {"type": "bool", "default": True},
                "role-write": {"type": "bool", "default": True},
            }
        },
    )
    # AND GIVEN that the container can connect
    state = base_state
    # WHEN we run any event
    state_out = ctx.run(ctx.on.update_status(), state)
    # THEN the charm sets Active: ready, even though we have no idea whether the endpoint is ready.
    assert state_out.unit_status == ops.ActiveStatus("read,write ready.")


def test_access_readiness_no_endpoint_raises():
    # GIVEN the caller doesn't pass an endpoint to Worker
    caller = MagicMock()
    with patch("cosl.juju_topology.JujuTopology.from_charm"):
        with patch("cosl.coordinated_workers.worker.Worker._reconcile"):
            worker = Worker(
                caller,
                "workload",
                lambda _: ops.pebble.Layer({"services": {"foo": {"command": "foo"}}}),
                {"cluster": "cluster"},
            )

    # THEN calling .check_readiness raises
    with pytest.raises(NoReadinessCheckEndpointConfiguredError):
        worker.check_readiness()  # noqa


def test_status_check_ready_with_patch(ctx, base_state, tls):
    with endpoint_ready(tls):
        with config_on_disk():
            with k8s_patch(status=ops.WaitingStatus("waiting")):
                state_out = ctx.run(ctx.on.config_changed(), base_state)
                assert state_out.unit_status == ops.WaitingStatus("waiting")
                with k8s_patch(status=ops.ActiveStatus("")):
                    state_out_out = ctx.run(ctx.on.update_status(), state_out)
                    assert state_out_out.unit_status == ops.ActiveStatus("read,write ready.")



import logging
import tempfile

import ops
import pytest
from ops import testing

from src.cosl.coordinated_workers.nginx import (
    CA_CERT_PATH,
    CERT_PATH,
    KEY_PATH,
    NGINX_CONFIG,
    Nginx,
)

logger = logging.getLogger(__name__)


@pytest.fixture
def certificate_mounts():
    temp_files = {}
    for path in {KEY_PATH, CERT_PATH, CA_CERT_PATH}:
        temp_file = tempfile.NamedTemporaryFile(delete=False)
        temp_files[path] = temp_file

    mounts = {}
    for cert_path, temp_file in temp_files.items():
        mounts[cert_path] = testing.Mount(location=cert_path, source=temp_file.name)

    # TODO: Do we need to clean up the temp files since delete=False was set?
    return mounts


@pytest.fixture
def nginx_context():
    return testing.Context(
        ops.CharmBase, meta={"name": "foo", "containers": {"nginx": {"type": "oci-image"}}}
    )


def test_certs_on_disk(certificate_mounts: dict, nginx_context: testing.Context):
    # GIVEN any charm with a container
    ctx = nginx_context

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=testing.State(
            containers={testing.Container("nginx", can_connect=True, mounts=certificate_mounts)}
        ),
    ) as mgr:
        charm = mgr.charm
        nginx = Nginx(charm, lambda: "foo_string", None)

        # THEN the certs exist on disk
        assert nginx.are_certificates_on_disk


def test_certs_deleted(certificate_mounts: dict, nginx_context: testing.Context):
    # Test deleting the certificates.

    # GIVEN any charm with a container
    ctx = nginx_context

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=testing.State(
            containers={testing.Container("nginx", can_connect=True, mounts=certificate_mounts)}
        ),
    ) as mgr:
        charm = mgr.charm
        nginx = Nginx(charm, lambda: "foo_string", None)

        # AND when we call delete_certificates
        nginx.delete_certificates()

        # THEN the certs get deleted from disk
        assert not nginx.are_certificates_on_disk


def test_reload_calls_nginx_binary_successfully(nginx_context: testing.Context):
    # Test that the reload method calls the nginx binary without error.

    # GIVEN any charm with a container
    ctx = nginx_context

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=testing.State(
            containers={
                testing.Container(
                    "nginx",
                    can_connect=True,
                    execs={testing.Exec(("nginx", "-s", "reload"), return_code=0)},
                )
            },
        ),
    ) as mgr:
        charm = mgr.charm
        nginx = Nginx(charm, lambda: "foo_string", None)

        # AND when we call reload
        # THEN the nginx binary is used rather than container restart
        assert nginx.reload() is None


def test_has_config_changed(nginx_context: testing.Context):
    # Test changing the nginx config and catching the change.

    # GIVEN any charm with a container and a nginx config file
    test_config = tempfile.NamedTemporaryFile(delete=False, mode="w+")
    ctx = nginx_context
    # AND when we write to the config file
    with open(test_config.name, "w") as f:
        f.write("foo")

    # WHEN we process any event
    with ctx(
        ctx.on.update_status(),
        state=testing.State(
            containers={
                testing.Container(
                    "nginx",
                    can_connect=True,
                    mounts={
                        "config": testing.Mount(location=NGINX_CONFIG, source=test_config.name)
                    },
                )
            },
        ),
    ) as mgr:
        charm = mgr.charm
        nginx = Nginx(charm, lambda: "foo_string", None)

        # AND a unique config is added
        new_config = "bar"

        # THEN the _has_config_changed method correctly determines that foo != bar
        assert nginx._has_config_changed(new_config)



# Copyright 2022 Canonical Ltd.
# See LICENSE file for licensing details.
"""## Overview.

This document explains how to use the `JujuTopology` class to
create and consume topology information from Juju in a consistent manner.

The goal of the Juju topology is to uniquely identify a piece
of software running across any of your Juju-managed deployments.
This is achieved by combining the following four elements:

- Model name
- Model UUID
- Application name
- Unit identifier


For a more in-depth description of the concept, as well as a
walk-through of it's use-case in observability, see
[this blog post](https://juju.is/blog/model-driven-observability-part-2-juju-topology-metrics)
on the Juju blog.

## Library Usage

This library may be used to create and consume `JujuTopology` objects.
The `JujuTopology` class provides three ways to create instances:

### Using the `from_charm` method

Enables instantiation by supplying the charm as an argument. When
creating topology objects for the current charm, this is the recommended
approach.

```python
topology = JujuTopology.from_charm(self)
```

### Using the `from_dict` method

Allows for instantion using a dictionary of relation data, like the
`scrape_metadata` from Prometheus or the labels of an alert rule. When
creating topology objects for remote charms, this is the recommended
approach.

```python
scrape_metadata = json.loads(relation.data[relation.app].get("scrape_metadata", "{}"))
topology = JujuTopology.from_dict(scrape_metadata)
```

### Using the class constructor

Enables instantiation using whatever values you want. While this
is useful in some very specific cases, this is almost certainly not
what you are looking for as setting these values manually may
result in observability metrics which do not uniquely identify a
charm in order to provide accurate usage reporting, alerting,
horizontal scaling, or other use cases.

```python
topology = JujuTopology(
    model="some-juju-model",
    model_uuid="00000000-0000-0000-0000-000000000001",
    application="fancy-juju-application",
    unit="fancy-juju-application/0",
    charm_name="fancy-juju-application-k8s",
)
```

"""
from collections import OrderedDict
from typing import Dict, List, Optional
from uuid import UUID

from ops.charm import CharmBase


class InvalidUUIDError(Exception):
    """Invalid UUID was provided."""

    def __init__(self, uuid: str):
        self.message = "'{}' is not a valid UUID.".format(uuid)
        super().__init__(self.message)


class JujuTopology:
    """JujuTopology is used for storing, generating and formatting juju topology information."""

    def __init__(
        self,
        model: str,
        model_uuid: str,
        application: str,
        unit: Optional[str] = None,
        charm_name: Optional[str] = None,
    ):
        """Build a JujuTopology object.

        A `JujuTopology` object is used for storing and transforming
        Juju topology information. This information is used to
        annotate Prometheus scrape jobs and alert rules. Such
        annotation when applied to scrape jobs helps in identifying
        the source of the scrapped metrics. On the other hand when
        applied to alert rules topology information ensures that
        evaluation of alert expressions is restricted to the source
        (charm) from which the alert rules were obtained.

        Args:
            model: a string name of the Juju model
            model_uuid: a globally unique string identifier for the Juju model
            application: an application name as a string
            unit: a unit name as a string
            charm_name: name of charm as a string
        """
        if not self.is_valid_uuid(model_uuid):
            raise InvalidUUIDError(model_uuid)

        self._model = model
        self._model_uuid = model_uuid
        self._application = application
        self._charm_name = charm_name
        self._unit = unit

    def is_valid_uuid(self, uuid: str) -> bool:
        """Validate the supplied UUID against the Juju Model UUID pattern.

        Args:
            uuid: string that needs to be checked if it is valid v4 UUID.

        Returns:
            True if parameter is a valid v4 UUID, False otherwise.
        """
        try:
            return str(UUID(uuid, version=4)) == uuid
        except (ValueError, TypeError):
            return False

    @classmethod
    def from_charm(cls, charm: CharmBase):
        """Creates a JujuTopology instance by using the model data available on a charm object.

        Args:
            charm: a `CharmBase` object for which the `JujuTopology` will be constructed
        Returns:
            a `JujuTopology` object.
        """
        return cls(
            model=charm.model.name,
            model_uuid=charm.model.uuid,
            application=charm.model.app.name,
            unit=charm.model.unit.name,
            charm_name=charm.meta.name,
        )

    @classmethod
    def from_dict(cls, data: Dict[str, str]):
        """Factory method for creating `JujuTopology` children from a dictionary.

        Args:
            data: a dictionary with five keys providing topology information. The keys are
                - "model"
                - "model_uuid"
                - "application"
                - "unit"
                - "charm_name"
                `unit` and `charm_name` may be empty, but will result in more limited
                labels. However, this allows us to support charms without workloads.

        Returns:
            a `JujuTopology` object.
        """
        return cls(
            model=data["model"],
            model_uuid=data["model_uuid"],
            application=data["application"],
            unit=data.get("unit", ""),
            charm_name=data.get("charm_name", ""),
        )

    def as_dict(
        self,
        *,
        remapped_keys: Optional[Dict[str, str]] = None,
        included_keys: Optional[List[str]] = None,
        excluded_keys: Optional[List[str]] = None,
    ) -> "OrderedDict[str, str]":
        """Format the topology information into an ordered dict.

        Keeping the dictionary ordered is important to be able to
        compare dicts without having to resort to deep comparisons.

        Args:
            remapped_keys: A dictionary mapping old key names to new key names,
                which will be substituted when invoked.
            included_keys: A list of key names to include in the returned dict.
                `excluded_keys` wins over `included_keys`.
            excluded_keys: A list of key names to exclude from the returned dict.
            uuid_length: The length to crop the UUID to.
        """
        ret = OrderedDict(
            [
                ("model", self.model),
                ("model_uuid", self.model_uuid),
                ("application", self.application),
                ("unit", self.unit),
                ("charm_name", self.charm_name),
            ]
        )

        if included_keys:
            ret = OrderedDict({k: v for k, v in ret.items() if k in included_keys})

        if excluded_keys:
            ret = OrderedDict({k: v for k, v in ret.items() if k not in excluded_keys})

        if remapped_keys:
            ret = OrderedDict(
                (remapped_keys.get(k), v) if remapped_keys.get(k) else (k, v) for k, v in ret.items()  # type: ignore
            )

        return ret  # type: ignore

    @property
    def identifier(self) -> str:
        """Format the topology information into a terse string.

        This crops the model UUID, making it unsuitable for comparisons against
        anything but other identifiers. Mainly to be used as a display name or file
        name where long strings might become an issue.

        >>> JujuTopology( \
              model = "a-model", \
              model_uuid = "00000000-0000-4000-8000-000000000000", \
              application = "some-app", \
              unit = "some-app/1" \
            ).identifier
        'a-model_00000000_some-app'
        """
        parts = self.as_dict(
            excluded_keys=["unit", "charm_name"],
        )

        parts["model_uuid"] = self.model_uuid_short
        values = parts.values()

        return "_".join([str(val) for val in values]).replace("/", "_")

    @property
    def label_matcher_dict(self) -> Dict[str, str]:
        """Format the topology information into a dict with keys having 'juju_' as prefix.

        Relabelled topology never includes the unit as it would then only match
        the leader unit (i.e. the unit that produced the dict).
        """
        items = self.as_dict(
            remapped_keys={"charm_name": "charm"},
            excluded_keys=["unit"],
        ).items()

        return {"juju_{}".format(key): value for key, value in items if value}

    @property
    def label_matchers(self) -> str:
        """Format the topology information into a promql/logql label matcher string.

        Topology label matchers should never include the unit as it
        would then only match the leader unit (i.e. the unit that
        produced the matchers).
        """
        items = self.label_matcher_dict.items()
        return ", ".join(['{}="{}"'.format(key, value) for key, value in items if value])

    @property
    def alert_expression_dict(self) -> Dict[str, str]:
        """Topology labels to insert in alert rule expressions.

        - "unit" is excluded because alert rules are forwarded over app data (one copy per app),
          so having a "unit" matcher would exclude alerts from all other units.
        - "charm" is excluded because in the case of subordinate charms, the labels need to match
          the subordinate and the principal.
        """
        items = self.as_dict(
            included_keys=["model", "model_uuid", "application"],
        ).items()

        return {"juju_{}".format(key): value for key, value in items if value}

    @property
    def alert_expression_str(self) -> str:
        """Topology label matchers for the alert expression, as a promql/logql string."""
        items = self.alert_expression_dict.items()
        return ", ".join(['{}="{}"'.format(key, value) for key, value in items if value])

    @property
    def model(self) -> str:
        """Getter for the juju model value."""
        return self._model

    @property
    def model_uuid(self) -> str:
        """Getter for the juju model uuid value."""
        return self._model_uuid

    @property
    def model_uuid_short(self) -> str:
        """Getter for the juju model value, truncated to the first eight letters."""
        return self._model_uuid[:8]

    @property
    def application(self) -> str:
        """Getter for the juju application value."""
        return self._application

    @property
    def charm_name(self) -> Optional[str]:
        """Getter for the juju charm name value."""
        return self._charm_name

    @property
    def unit(self) -> Optional[str]:
        """Getter for the juju unit value."""
        return self._unit



# Copyright 2020 Canonical Ltd.
# See LICENSE file for licensing details.

import unittest
import unittest.mock
from pathlib import PosixPath

from cosl import CosTool


class TestTool(unittest.TestCase):
    """Test that the cos-tool base implementation works."""

    def assert_type_error_helper(self, func, *args, **kwargs):
        with self.assertRaises(TypeError) as cm:
            func(*args, **kwargs)
        self.assertIn("Either a default query", str(cm.exception))

    # pylint: disable=protected-access
    @unittest.mock.patch("platform.machine", lambda: "teakettle")
    def test_disable_on_invalid_arch(self):
        tool = CosTool(default_query_type="logql")
        self.assertIsNone(tool.path)
        self.assertTrue(tool._disabled)

    # pylint: disable=protected-access
    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_gives_path_on_valid_arch(self):
        """When given a valid arch, it should return the binary path."""
        tool = CosTool(default_query_type="promql")
        self.assertIsInstance(tool.path, PosixPath)

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_setup_transformer(self):
        """When setup it should know the path to the binary."""
        tool = CosTool(default_query_type="promql")

        self.assertIsInstance(tool.path, PosixPath)

        p = str(tool.path)
        self.assertTrue(p.endswith("cos-tool-amd64"))

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_typeerror_is_raised_if_no_query_is_used(self):
        """If no default query type or querytpye is set, it should raise."""
        tool = CosTool()

        self.assert_type_error_helper(tool.apply_label_matchers, rules={})
        self.assert_type_error_helper(tool.validate_alert_rules, rules={})
        self.assert_type_error_helper(tool.inject_label_matchers, expression="", topology={})

        p = str(tool.path)
        self.assertTrue(p.endswith("cos-tool-amd64"))



# Copyright 2020 Canonical Ltd.
# See LICENSE file for licensing details.

import subprocess
import unittest
import unittest.mock

from cosl import CosTool


class TestTransformLogQL(unittest.TestCase):
    """Test that the cos-tool logql implementation works."""

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    @unittest.mock.patch("subprocess.run")
    def test_returns_original_expression_when_subprocess_call_errors(self, mocked_run):
        mocked_run.side_effect = subprocess.CalledProcessError(
            returncode=10, cmd="cos-tool", stderr=""
        )

        tool = CosTool(default_query_type="logql")
        output = tool.apply_label_matchers(
            {
                "groups": [
                    {
                        "alert": "CPUOverUse",
                        "expr": '{job="foo"} |= "info"',
                        "for": "0m",
                        "labels": {
                            "severity": "Low",
                            "juju_model": "None",
                            "juju_model_uuid": "f2c1b2a6-e006-11eb-ba80-0242ac130004",
                            "juju_application": "consumer-tester",
                        },
                        "annotations": {
                            "summary": "Instance {{ $labels.instance }} CPU over use",
                            "description": "{{ $labels.instance }} of job "
                            "{{ $labels.job }} has used too much CPU.",
                        },
                    }
                ]
            }
        )
        self.assertEqual(output["groups"][0]["expr"], '{job="foo"} |= "info"')

    @unittest.mock.patch("platform.machine", lambda: "invalid")
    def test_uses_original_expression_when_binary_missing(self):
        tool = CosTool(default_query_type="logql")
        output = tool.apply_label_matchers(
            {
                "groups": [
                    {
                        "alert": "CPUOverUse",
                        "expr": '{job="foo"} |= "info"',
                        "for": "0m",
                        "labels": {
                            "severity": "Low",
                            "juju_model": "None",
                            "juju_model_uuid": "f2c1b2a6-e006-11eb-ba80-0242ac130004",
                            "juju_application": "consumer-tester",
                        },
                        "annotations": {
                            "summary": "Instance {{ $labels.instance }} CPU over use",
                            "description": "{{ $labels.instance }} of job "
                            "{{ $labels.job }} has used too much CPU.",
                        },
                    }
                ]
            }
        )
        self.assertEqual(output["groups"][0]["expr"], '{job="foo"} |= "info"')

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_fetches_the_correct_expression(self):
        tool = CosTool(default_query_type="logql")

        output = tool.inject_label_matchers(
            '{env="production"}', {"juju_model": "some_juju_model"}
        )
        assert output == '{env="production", juju_model="some_juju_model"}'

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_handles_comparisons(self):
        tool = CosTool(default_query_type="logql")
        output = tool.inject_label_matchers(
            'rate({env="production"} |= "info" [10m]) > 1', {"juju_model": "some_juju_model"}
        )
        assert (
            output == '(rate({env="production", juju_model="some_juju_model"} |= "info"[10m]) > 1)'
        )

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_handles_multiple_labels(self):
        tool = CosTool(default_query_type="logql")
        keys = {
            "juju_model": "some_juju_model",
            "juju_model_uuid": "123ABC",
            "juju_application": "some_application",
            "juju_unit": "some_application/1",
        }
        output = tool.inject_label_matchers('{env="production"}', keys)
        self.assertTrue(all('{}="{}"'.format(k, v) in output for k, v in keys.items()))


class TestValidateAlertsLogQL(unittest.TestCase):
    """Test that the cos-tool logql validation works."""

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_returns_errors_on_bad_rule_file(self):
        tool = CosTool(default_query_type="logql")
        valid, errs = tool.validate_alert_rules(
            {
                "groups": [
                    {
                        "alert": "BadSyntax",
                        "expr": "rate{) > 0.12",
                    }
                ]
            }
        )
        self.assertEqual(valid, False)
        self.assertIn("error validating", errs)

    @unittest.mock.patch("platform.machine", lambda: "x86_64")
    def test_successfully_validates_good_alert_rules(self):
        tool = CosTool(default_query_type="logql")
        valid, errs = tool.validate_alert_rules(
            {
                "groups": [
                    {
                        "alert": "CPUOverUse",
                        "expr": 'rate({job="unit_test"} [5m]) > 0.12',
                        "for": "0m",
                        "labels": {
                            "severity": "Low",
                            "juju_model": "None",
                            "juju_model_uuid": "f2c1b2a6-e006-11eb-ba80-0242ac130004",
                            "juju_application": "consumer-tester",
                        },
                        "annotations": {
                            "summary": "Instance {{ $labels.instance }} CPU over use",
                            "description": "{{ $labels.instance }} of job "
                            "{{ $labels.job }} has used too much CPU.",
                        },
                    }
                ]
            }
        )
        self.assertEqual(errs, "")
        self.assertEqual(valid, True)



# Copyright 2023 Canonical Ltd.
# See LICENSE file for licensing details.

"""Grafana Dashboard."""

import base64
import binascii
import hashlib
import json
import logging
import lzma
import warnings
from typing import Any, ClassVar, Dict, Tuple, Union

logger = logging.getLogger(__name__)


class GrafanaDashboard(str):
    """GrafanaDashboard represents an actual dashboard in Grafana.

    The class is used to compress and encode, or decompress and decode,
    Grafana Dashboards in JSON format using LZMA.
    """

    @staticmethod
    def _serialize(raw_json: Union[str, bytes]) -> "GrafanaDashboard":
        warnings.warn(
            "GrafanaDashboard._serialize is deprecated; use LZMABase64.compress(json.dumps(...)) instead.",
            category=DeprecationWarning,
        )
        return GrafanaDashboard(LZMABase64.compress(raw_json))

    def _deserialize(self) -> Dict[str, Any]:
        warnings.warn(
            "GrafanaDashboard._deserialize is deprecated; use json.loads(LZMABase64.decompress(...)) instead.",
            category=DeprecationWarning,
        )
        try:
            return json.loads(LZMABase64.decompress(self))
        except json.decoder.JSONDecodeError as e:
            logger.error("Invalid Dashboard format: %s", e)
            return {}

    def __repr__(self):
        """Return string representation of self."""
        return "<GrafanaDashboard>"


class LZMABase64:
    """A helper class for LZMA-compressed-base64-encoded strings.

    This is useful for transferring over juju relation data, which can only have keys of type string.
    """

    @classmethod
    def compress(cls, raw_json: Union[str, bytes]) -> str:
        """LZMA-compress and base64-encode into a string."""
        if not isinstance(raw_json, bytes):
            raw_json = raw_json.encode("utf-8")
        return base64.b64encode(lzma.compress(raw_json)).decode("utf-8")

    @classmethod
    def decompress(cls, compressed: str) -> str:
        """Decompress from base64-encoded-lzma-compressed string."""
        return lzma.decompress(base64.b64decode(compressed.encode("utf-8"))).decode()


class DashboardPath40UID:
    """A helper class for dashboard UID of length 40, generated from charm name and dashboard path."""

    length: ClassVar[int] = 40

    @classmethod
    def _hash(cls, components: Tuple[str, ...], length: int) -> str:
        return hashlib.shake_256("-".join(components).encode("utf-8")).hexdigest(length)

    @classmethod
    def generate(cls, charm_name: str, dashboard_path: str) -> str:
        """Generate a dashboard uid from charm name and dashboard path.

        The combination of charm name and dashboard path (relative to the charm root) is guaranteed to be unique across
        the ecosystem. By design, this intentionally does not take into account instances of the same charm with
        different charm revisions, which could have different dashboard versions.
        Ref: https://github.com/canonical/observability/pull/206

        The max length grafana allows for a dashboard uid is 40.
        Ref: https://grafana.com/docs/grafana/latest/developers/http_api/dashboard/#identifier-id-vs-unique-identifier-uid

        Args:
            charm_name: The name of the charm (not app!) that owns the dashboard.
            dashboard_path: Path (relative to charm root) to the dashboard file.

        Returns: A uid based on the input args.
        """
        # Since the digest is bytes, we need to convert it to a charset that grafana accepts.
        # Let's use hexdigest, which means 2 chars per byte, reducing our effective digest size to 20.
        return cls._hash((charm_name, dashboard_path), cls.length // 2)

    @classmethod
    def is_valid(cls, uid: str) -> bool:
        """Check if the given UID is a valid "Path-40" UID.

        The UID must be of a particular length, and since we generate it with hexdigest() then we also know it must
        unhexlify.

        This is not a bullet-proof check, because it's plausible that some dashboard would have a 40-length hexstring as
        its uid, but given the current state of the ecosystem, it's quite unlikely.
        """
        if not uid:
            return False
        if len(uid) != cls.length:
            return False
        try:
            binascii.unhexlify(uid)
        except binascii.Error:
            return False
        return True



#!/usr/bin/env python3
# Copyright 2024 Canonical
# See LICENSE file for licensing details.

"""Shared utilities for the cosl interfaces."""

import json
import logging
from typing import (
    MutableMapping,
    Optional,
)

import pydantic
from pydantic import ConfigDict

log = logging.getLogger("utils")

BUILTIN_JUJU_KEYS = {"ingress-address", "private-address", "egress-subnets"}

# =================
# | Databag Model |
# =================

# Note: MutableMapping is imported from the typing module and not collections.abc
# because subscripting collections.abc.MutableMapping was added in python 3.9, but
# most of our charms are based on 20.04, which has python 3.8.

_RawDatabag = MutableMapping[str, str]


class DataValidationError(Exception):
    """Raised when relation databag validation fails."""


class DatabagModel(pydantic.BaseModel):
    """Base databag model."""

    model_config = ConfigDict(
        # tolerate additional keys in databag
        extra="ignore",
        # Allow instantiating this class by field name (instead of forcing alias).
        populate_by_name=True,
    )  # type: ignore
    """Pydantic config."""

    @classmethod
    def load(cls, databag: _RawDatabag):
        """Load this model from a Juju databag."""
        try:
            data = {
                k: json.loads(v)
                for k, v in databag.items()
                # Don't attempt to parse model-external values
                if k in {(f.alias or n) for n, f in cls.__fields__.items()}  # type: ignore
            }
        except json.JSONDecodeError as e:
            msg = f"invalid databag contents: expecting json. {databag}"
            log.error(msg)
            raise DataValidationError(msg) from e

        try:
            return cls.model_validate_json(json.dumps(data))  # type: ignore
        except pydantic.ValidationError as e:
            msg = f"failed to validate databag: {databag}"
            if databag:
                log.debug(msg, exc_info=True)
            raise DataValidationError(msg) from e

    def dump(self, databag: Optional[_RawDatabag] = None, clear: bool = True) -> _RawDatabag:
        """Write the contents of this model to Juju databag.

        :param databag: the databag to write the data to.
        :param clear: ensure the databag is cleared before writing it.
        """
        _databag: _RawDatabag = {} if databag is None else databag

        if clear:
            _databag.clear()

        dct = self.model_dump(mode="json", by_alias=True, exclude_defaults=True)  # type: ignore
        _databag.update({k: json.dumps(v) for k, v in dct.items()})
        return _databag


# FIXME: in pydantic v2, the json stuff we've been doing is no longer necessary.
#  It becomes much easier to work with Json fields and the databagmodel class becomes much simpler.
#  We should rewrite the cluster implementation to use this class,
#  and replace the original DatabagModel with it
class DatabagModelV2(pydantic.BaseModel):
    """Base databag model."""

    model_config = ConfigDict(
        # tolerate additional keys in databag
        extra="ignore",
        # Allow instantiating this class by field name (instead of forcing alias).
        populate_by_name=True,
    )  # type: ignore
    """Pydantic config."""

    @classmethod
    def load(cls, databag: _RawDatabag):
        """Load this model from a Juju databag."""
        try:
            return cls.model_validate_json(json.dumps(dict(databag)))  # type: ignore
        except pydantic.ValidationError as e:
            msg = f"failed to validate databag: {databag}"
            if databag:
                log.debug(msg, exc_info=True)
            raise DataValidationError(msg) from e

    def dump(self, databag: Optional[_RawDatabag] = None, clear: bool = True) -> _RawDatabag:
        """Write the contents of this model to Juju databag.

        :param databag: the databag to write the data to.
        :param clear: ensure the databag is cleared before writing it.
        """
        _databag: _RawDatabag = {} if databag is None else databag

        if clear:
            _databag.clear()

        dct = self.model_dump(mode="json", by_alias=True, exclude_defaults=True, round_trip=True)  # type: ignore
        _databag.update(dct)
        return _databag
